//
// CfgParser.hh
// Copyright (C) 2002-2005 Claes Nasten <pekdon{@}pekdon{.}net>
//                         Alaa Abd El Fatah <alaa at linux-egypt dot org>
//
// This program is licensed under the GNU GPL.
// See the LICENSE file for more information.
//

#include "../config.h"

#include "CfgParser.hh"
#include "Util.hh"

#include <cstdio>
#include <iostream>
#include <functional>
#include <sstream>
#include <fstream>
#include <map>

extern "C" {
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include <signal.h>
}

using std::cerr;
using std::endl;
using std::string;
using std::ifstream;
using std::istream;
using std::stringstream;
using std::list;
using std::map;
using std::find;

namespace Parser {

//! @brief substr wrapper, with bounds checking
inline string
safeSubstr(const string &str, string::size_type s, string::size_type e)
{
	if ((s > str.size()) || ((s + e) > str.size())) {
		return string("");
	} else {
		return str.substr(s, e);
	}
}

//! @brief Executes command and fills dest with the output.
inline void
pipeFillBuffer(stringstream &dest, const string &command)
{
	int fd[2];

	if (pipe(fd) == -1) {
		cerr << __FILE__ << "@" << __LINE__ << ": "
				 << "pipeFillBuffer: pipe() failed" << endl;
		return;
	}

	// remove signal handler ( otherwise this doesn't work full time )
	struct sigaction act, oldact;
        
	act.sa_handler = SIG_DFL;
	act.sa_mask = sigset_t();
	act.sa_flags = 0;
	sigaction(SIGCHLD, &act, &oldact);

	pid_t pid = fork();
	if (pid == -1) { // error
		cerr << __FILE__ << "@" << __LINE__ << ": "
				 << "pipeFillBuffer: fork() failed" << endl;

	} else if (pid == 0) { // child process
		dup2(fd[1], STDOUT_FILENO);	// stdout now points to the pipe

		close(fd[0]);	// close unused filedescriptor
		close(fd[1]);	// close unused filedescriptor

		execlp("/bin/sh", "sh", "-c", command.c_str(), (char *) NULL);

		cerr << __FILE__ << "@" << __LINE__ << ": "
				 << "pipeFillBuffer: execlp() failed" << endl;

		close(STDOUT_FILENO);

		exit(1);

	} else { // parent process
		close(fd[1]);	// close unused filedescriptor

		FILE *source = fdopen(fd[0],"r");
		if (source) {
			int c;
			while (!feof(source) && !ferror(source) && ((c = fgetc(source)) != EOF))
				dest.put((char) c);

			fclose(source);
		}

		int status;
		if (waitpid(pid, &status, 0) == -1) {
			cerr << __FILE__ << "@" << __LINE__ << ": "
					 << "pipeFillBuffer: waitpid failed" << endl;
		}
	}

	// restore signal handler
	sigaction(SIGCHLD, &oldact, NULL);
}

//! @brief Appends content of source until delim
inline void
updateBuffer(stringstream &dest, istream &source, const char delim)
{
	source.get(*dest.rdbuf(), delim);

#if ( __GNUC__ == 2 ) && ( __GNUC_MINOR__ == 96 )
	dest.seekg(0, std::ios::beg);
#endif
	dest.clear();
}

//! @brief Replaces content of buf with str
inline void
updateBuffer(stringstream &buf, const string &str) {
#if (__GNUC__ == 2) && (__GNUC_MINOR__ == 95)
	buf.str("");
	buf.clear();
	buf << str;
#elif (__GNUC__ == 2) && (__GNUC_MINOR__ == 96)
	buf.str(str);
	buf.seekg(0, std::ios::beg);
#else
	buf.str(str);
#endif
	buf.clear();
}

//! @brief Skips to the end of a comment.
inline void
parseComment(stringstream &buf, const string::size_type begin, istream &file)
{
	string::size_type end = buf.str().find_first_of('\n', begin);
	if (end != string::npos) {
		updateBuffer(buf, Parser::safeSubstr(buf.str(), end + 1,
																				 buf.str().size() - end -1));
	} else {
		updateBuffer(buf, file, '\n');
		buf.str("");
	}
}

//! @brief Skips to the end of the line ( C++ ) or until next */ ( C )
inline void
parseCComment(stringstream &buf, const string::size_type begin, istream &file)
{
	// get the next char, comment if / or *
	char next;
	if (begin == (buf.str().size() - 1)) {
		next = file.peek();
	} else {
		next = buf.str()[begin + 1];
	}

	// skip C++ comment
	if (next == '/') {
		parseComment(buf, begin, file);

	// skip C comment
	} else if (next == '*') {
		string::size_type pos = buf.str().find("*/", begin);
		if (pos != string::npos) {
			updateBuffer(buf, Parser::safeSubstr(buf.str(), pos + 2,
																					 buf.str().size() - pos - 2));

		} else {
			for (bool exit = false; !file.eof() && (exit != true); ) {
				updateBuffer(buf, file, '/');
				file.ignore(); // skip the /

				// check for the end of the comment
				if (buf.str()[buf.str().size() - 1] == '*') {
					exit = true;
				}

				updateBuffer(buf, ""); // flush the buffer, no need to eat memory
			}
		}
	// not a C or C++ comment, garbage?! skip character
	} else {
		updateBuffer(buf, Parser::safeSubstr(buf.str(), begin + 1,
																				 buf.str().size() - begin - 1));
	}
}

//! @brief Gets the name in a Name = "Value" stream
inline void
parseName(stringstream &buf, string::size_type end, string &name)
{
	string::size_type pos, begin = 0;

	for (pos = end - 1; pos >= 0; --pos) {
		if ((buf.str()[pos] != '=') && (isspace(buf.str()[pos]) == 0)) {
			end = pos + 1;
			break;
		}
	}

	for (pos = end - 1; pos >= 0; --pos) {
		if ((buf.str()[pos] != '_') && (isalnum(buf.str()[pos]) == 0)) {
			begin = pos + ((buf.str()[pos] == '$') ? 0 : 1);
			break;
		}
	}

	name = buf.str().substr(begin, end - begin);
}

//! @brief Gets the value in a " pair
inline void
parseValue(stringstream &buf, const string::size_type begin, string &value)
{
	string::size_type esc = begin + 1;
	string::size_type end = buf.str().find_first_of("\n\"", begin + 1);
	value = "";

	// get until end of " pair , skip escaped newlines
	while ((end != string::npos) && (buf.str()[end - 1] == '\\')) {
		value += buf.str().substr(esc, end - esc - 1 );
		esc = end;
		end = buf.str().find_first_of("\n\"", end + 1);
	}

	if (end != string::npos) {
		value += buf.str().substr(esc, end - esc);
		updateBuffer(buf, Parser::safeSubstr(buf.str(), end + 1,
																				 buf.str().size() - end - 1));
	} else { //something went wrong
		value += buf.str().substr(esc, buf.str().size() - esc);
		buf.str("");
	}
}


//! @brief Adds value to the vars map, reflecting changes to environment
inline void
variableDefine(map<string, string> &vars,
							 const string &name, const string &value)
{
	vars[name] = value;

	// reflect changes to environment
	if ((name.size() > 2) && (name[1] == '_')) {
		setenv(name.c_str() + 2, value.c_str(), 1);
	}
}

//! @brief Replaces $VARIABLE in string
inline void
variableReplace(map<string, string> &vars, string &variable,
								const string::size_type begin, const string::size_type end)
{
	// if the variable starts with _ treat it as a global variale from
	// the environment.
	if ((end > (begin + 2)) && (variable[begin + 1] == '_')) {
		char *env = getenv(variable.substr(begin + 2, end - begin - 2).c_str());
		variable.replace(begin, end, (env != NULL) ? env : "");

	} else {
		variable.replace(begin, end - begin,
										 vars[variable.substr(begin, end - begin)]);
	}
}

//! @brief Expands and replaces variables.
inline void
parseVariables(CfgParser::Item &temp_ci, map<string, string> &vars) {
	string::size_type begin = 0;
	string::size_type end;

	// expand variables
	while ((begin = temp_ci.value.find_first_of('$', begin)) != string::npos) {

		// escape backslash
		if (temp_ci.value[begin - 1] == '\\' ) {
			temp_ci.value.erase(begin - 1, 1);

		} else {

			// if global variable, skip the leading _
			end = begin + ((temp_ci.value[begin + 1] == '_' ) ? 2 : 1);

			// get the end of the comment ( first non alphanum character,
      // _ also accepted )
			for (; end != temp_ci.value.size(); ++end) {
				if ((isalnum(temp_ci.value[end]) == 0) &&
						(temp_ci.value[end] != '_')) {
					break;
				}
			}

			Parser::variableReplace(vars, temp_ci.value, begin, end);
		}
	}

	// define variable
	if (temp_ci.name[0] == '$') {
		Parser::variableDefine(vars, temp_ci.name, temp_ci.value);
	}
}

} // end namespace Parser

//! @brief Opens an ifstream for loading and loads
bool
CfgParser::Section::load(const std::string &filename)
{
	if (filename.size() == 0) {
		return false;
	}

	string filename_real(filename);

	// first try cwd or absolute
	ifstream file(filename_real.c_str());
	if (!file.is_open()) {
		file.close();
		file.clear();

		// If opening failes, try current source dir if we have one
		// and the filename is relative
		if ((_sources_list.size()) && (filename[0] != '/')) {
			filename_real = Util::getDir(_sources_list.back());
			filename_real += "/" + filename;

			file.open(filename_real.c_str());
		}
	}

	if (file.is_open()) {
		_sources_list.push_back(filename_real);
		load(file);
		file.close();
		_sources_list.pop_back();

		return true;
	}

	return false;
}

//! @brief Loads and parses an istream
void
CfgParser::Section::load(std::istream &file)
{
	string::size_type p;
	stringstream buf("");
	string temp_val;
	static string rest_of_buf = "";
	CfgParser::Item temp_ci;

	while(!file.eof() || rest_of_buf.size()) {
		if (rest_of_buf.size()) {
			Parser::updateBuffer(buf, rest_of_buf);
			rest_of_buf = "";
		} else {
		  Parser::updateBuffer(buf,file,'{');
#if ( __GNUC__ == 2 )
			if (!buf.str().size()) {
				_sect_it = _sect_list.begin();
				return;
			}
#endif //__GNUC__
		}
		while ((p = buf.str().find_first_of("\"}#/;")) != string::npos) {
			switch(buf.str()[p]) {
			case ';':
				Parser::updateBuffer(buf, Parser::safeSubstr(buf.str(), p + 1,
																										 buf.str().size() - p - 1));
				break;
			case '#': // comment
				Parser::parseComment(buf, p, file);
				break;
			case '/': // C or C++ comment
				Parser::parseCComment(buf, p, file);
				break;
			case '}':
				if (_name.size()) {  // close section
					_sect_it = _sect_list.begin();
					// return whats between } and { to the previous call
					rest_of_buf = Parser::safeSubstr(buf.str(), p + 1,
																					 buf.str().size() - p - 1);
					return;
				} else { // not end of scope this is just an extra '}'
					Parser::updateBuffer(buf, Parser::safeSubstr(buf.str(), p + 1,
																											 buf.str().size() - p - 1));
				}
				break;
			case '\"': // name/value pair
				Parser::parseName(buf, p, temp_ci.name);
				Parser::parseValue(buf, p, temp_ci.value);
				Parser::parseVariables(temp_ci, _variables);

				if (temp_ci.name == "INCLUDE") {
					Util::expandFileName(temp_ci.value);
					load(temp_ci.value);
				} else if (temp_ci.name == "COMMAND") {
					stringstream temp_buf("");
					Parser::pipeFillBuffer(temp_buf, temp_ci.value);
					load(temp_buf);
				} else if (temp_ci.name[0] != '$') {
					_values.push_back(temp_ci);
				}
				break;
			default:
				// do nothing
				break;
			}
		}

		// end of section.
		if (file.peek() == '{'){
			file.ignore(); // by_pass '{'

			buf.clear();
			temp_val = "";
			buf >> temp_val;
			if (temp_val.size()) {
				_sect_list.push_back(CfgParser::Section(_sources_list, _variables,
																								temp_val));
				buf.str("");
				_sect_list.back().load(file);
			} else if (!_values.empty()) {
				temp_ci = _values.back();
				_sect_list.push_back(CfgParser::Section(_sources_list, _variables,
																								temp_ci.name));
				temp_ci.name = "Name";
				_sect_list.back()._values.push_back(temp_ci);
				_values.pop_back();
				buf.str("");
				_sect_list.back().load(file);
			}
		}
	}

	rest_of_buf = "";
	_sect_it = _sect_list.begin();
}

//! @brief Gets the value name from the config_list
//! @return True on success else false.
bool
CfgParser::Section::getValue(const std::string &name, int &value)
{
	_it = find(_values.begin(), _values.end(), name);
	if (_it != _values.end()) {
		value = strtol(_it->value.c_str(), NULL, 10);
		_values.erase(_it);
		return true;
	}
	return false;
}

//! @brief Gets the value name from the config_list
//! @return True on success else false.
bool
CfgParser::Section::getValue(const std::string &name, uint &value)
{
	_it = find(_values.begin(), _values.end(), name);
	if (_it != _values.end()) {
		value = (unsigned) strtol(_it->value.c_str(), NULL, 10);
		_values.erase(_it);
		return true;
	}
	return false;
}

//! @brief Gets the value name from the config_list
//! @return True on success else false.
bool
CfgParser::Section::getValue(const std::string &name, bool &value)
{
	_it = find(_values.begin(), _values.end(), name);
	if (_it != _values.end()) {
		value = Util::isTrue(_it->value);
		return true;
	}
	return false;
}

//! @brief Gets the value name from the config_list
//! @return True on success else false.
bool
CfgParser::Section::getValue(const std::string &name, std::string &value)
{
	_it = find(_values.begin(), _values.end(), name);
	if (_it != _values.end()) {
		value = _it->value;
		_values.erase(_it);
		return true;
	}
	return false;
}

//! @brief Gets the next value from the config file.
//! @return True if any objects left in the config_list
bool
CfgParser::Section::getNextValue(std::string &name, std::string &value)
{
	if (_values.size()) {
		name = _values.front().name;
		value = _values.front().value;
		_values.pop_front();
		return true;
	}
	return false;
}

//! @brief Parses the file, unloads allready parsed info.
//! @param filename Name of file to parse.
bool
CfgParser::load(const std::string &filename)
{
	return _cfg.load(filename);
}

//! @brief Parses the output of command.
//! @param command Command to execute and parse output from.
bool
CfgParser::loadCommand(const std::string &command)
{
	if (command.size() == 0) {
		return false;
	}

	stringstream command_output("");
	Parser::pipeFillBuffer(command_output, command);
	_cfg.load(command_output);

	return true;
}
