//
// Image.cc for pekwm
// Copyright (C) 2003-2004 Claes Nasten <pekdon{@}pekdon{.}net>
//
// This program is licensed under the GNU GPL.
// See the LICENSE file for more information.
//

#include "../config.h"
#include "Image.hh"
#include "PScreen.hh"
#include "PDecor.hh"
#include "ScreenResources.hh"

#ifdef DEBUG
#include <iostream>
using std::cerr;
using std::endl;
#endif // DEBUG
using std::string;

//! @brief Image constructor
Image::Image(Display *dpy) :
_dpy(dpy),
_pixmap(None), _mask(None),
_width(0), _height(0), _image_type(IMAGE_TYPE_TILED)
{
#ifdef HAVE_IMLIB2
	_im_image = NULL;
#else // !HAVE_IMLIB2
	_xpm_image.data = NULL; // so that we can track if it's loaded or not
#endif // HAVE_IMLIB2
}

//! @brief Image destructor
Image::~Image(void)
{
	unload();
}

//! @brief Loads an xpm image into the image
//! @param file File to load
//! @return true if image was successfully loaded, else false.
bool
Image::load(const std::string &file)
{
	if (file.size() == 0) {
		return false;
	}

	// Unload allready loaded image
	unload();

	bool status = false;

#ifdef HAVE_IMLIB2
	imlib_context_push(ScreenResources::instance()->getImlibContext());

	_im_image = imlib_load_image(file.c_str());

	if (_im_image != NULL) {
		imlib_context_set_image(_im_image);

		imlib_render_pixmaps_for_whole_image(&_pixmap, &_mask);

		_width = imlib_image_get_width();
		_height = imlib_image_get_height();

		imlib_context_pop();

		status = true;
	}
#else // !HAVE_IMLIB2
	// Setup loading attributes attributes
	XpmAttributes attr;
	attr.valuemask = XpmSize;

	int load =
		XpmReadFileToPixmap(_dpy, PScreen::instance()->getRoot(), (char *) file.c_str(),
												&_pixmap, &_mask, &attr);

	// Image loaded okay, now let's convert it into a XImage and setup the sizes
	if (load == XpmSuccess) {
		_width = attr.width;
		_height = attr.height;

		// move the pixmap into a xpmimage
		XpmCreateXpmImageFromPixmap(_dpy, _pixmap, _mask, &_xpm_image, NULL);
		status = true;
	}
#endif // HAVE_IMLIB2

	return status;
}

//! @brief Unloads the image
void
Image::unload(void)
{
#ifdef HAVE_IMLIB2
	if (_im_image != NULL) {
		imlib_context_push(ScreenResources::instance()->getImlibContext());

		imlib_context_set_image(_im_image);

		imlib_free_image_and_decache();
		imlib_free_pixmap_and_mask(_pixmap);

		imlib_context_pop();

		_im_image = NULL;
	}
#else // !HAVE_IMLIB_2
	if (_xpm_image.data != NULL) {
		XpmFreeXpmImage(&_xpm_image);
		_xpm_image.data = NULL;
	}
	if (_pixmap != None) {
		XFreePixmap(_dpy, _pixmap);
		_pixmap = None;
	}
	if (_mask != None) {
		XFreePixmap(_dpy, _mask);
		_mask = None;
	}
#endif // HAVE_IMLIB2

	_width = _height = 0;
}

//! @brief Renders the image on the drawable dest
void
Image::render(Drawable dest, int x, int y, uint width, uint height)
{
	// init rendering
#ifdef HAVE_IMLIB2
	imlib_context_push(ScreenResources::instance()->getImlibContext());

	imlib_context_set_image(_im_image);
	imlib_context_set_drawable(dest);
#else // !HAVE_IMLIB2
	if (_xpm_image.data == NULL) {
		return;
	}
#endif // HAVE_IMLIB2

	// can't have 0 wide or tall thing
	if (width == 0) {
		width = _width;
	}
	if (height == 0) {
		height = _height;
	}

	// scaled image rendering differs from tiled
	if (_image_type == IMAGE_TYPE_SCALED) {
#ifdef HAVE_IMLIB2
		imlib_render_image_on_drawable_at_size(x, y, width, height);
#else // !HAVE_IMLIB2
		Image *image = getScaled(width, height);

		if (image != NULL) {
			XCopyArea(_dpy, image->_pixmap, dest, PScreen::instance()->getGC(),
								0, 0, width, height, x, y);
			delete image;
		}
#endif // HAVE_IMLIB2

	} else {
		XGCValues gv;
		gv.fill_style = FillTiled;
		gv.tile = _pixmap;
		gv.ts_x_origin = x; // to make the tile start at the correct position
		gv.ts_y_origin = y;

		GC gc = XCreateGC(_dpy, dest,
											GCTile|GCFillStyle|GCTileStipXOrigin|GCTileStipYOrigin,
											&gv);
		XFillRectangle(_dpy, dest, gc, x, y, width, height);

		XFreeGC(_dpy, gc);
	}

#ifdef HAVE_IMLIB2
	imlib_context_pop();
#endif // HAVE_IMLIB2
}

//! @brief Scales the image permanent.
void
Image::scale(uint width, uint height)
{
	// FIXME: Permanent scaling
}

//! @brief Returns a scaled version of the Image
//! @todo  Put the Destination XpmImage back.
Image*
Image::getScaled(uint width, uint height)
{
#ifdef HAVE_IMLIB2
	return NULL;
#else // !HAVE_IMLIB2
	if (_xpm_image.data == NULL)
		return NULL;

	// setup destination image
	XpmImage dest;
	dest.width = width;
	dest.height = height;
	dest.cpp = _xpm_image.cpp;
	dest.ncolors = _xpm_image.ncolors;
	dest.colorTable = _xpm_image.colorTable;
	dest.data = new uint[width * height];

	// scale ratios
	float rx = float(_width) / float(width);
	float ry = float(_height) / float(height);

	// counters
	float sx, sy;
	uint x, y, loff = 0;

	uint *data = dest.data;
	// scale the image
	for (y = 0, sy = 0; y < height; ++y, sy += ry) {
		loff = ((int) sy) * _width;
		for (x = 0, sx = loff; x < width; ++x, sx += rx) {
			*data++ = _xpm_image.data[(int) sx];
		}
	}

	// now create the new scaled pixmap
	Image *image = new Image(_dpy);
	XpmCreatePixmapFromXpmImage(_dpy, PScreen::instance()->getRoot(), &dest,
															&image->_pixmap, &image->_mask, NULL);
	image->_width = width;
	image->_height = height;

	// free memory
	delete [] dest.data;
	return image;
#endif // HAVE_IMLIB2
}
